# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_html_parsing.ipynb (unless otherwise specified).

__all__ = ['get_file_paths', 'RE_HTM', 'collect_sheet_uris', 'RE_SHEET', 'download_sheet', 'download_multiple_sheets',
           'RE_FNAME', 'is_date', 'get_sheet_df', 'handle_title_and_date', 'disambiguate_party', 'VOTE_COLS',
           'PARTY_MAP', 'get_squished_dataframe', 'set_sheet_dtypes', 'DTYPES', 'get_multiple_sheets_df',
           'get_multiple_sheets']

# Cell
from bs4 import BeautifulSoup
from pathlib import Path
import re, os
import typing
import requests
import tqdm
import time
import pandas as pd
import itertools
from fastcore.all import *

# Cell
RE_HTM = re.compile('(\.html?)')

def get_file_paths(path:typing.Union[Path,str], suffix:str=None, pattern=None):
    'Collecting files with a specific suffix or pattern from `path`'
    return Path(path).ls().filter(lambda x: x.suffix=='.htm' or (pattern and pattern.search(str(x)))).unique()

# Cell
RE_SHEET = re.compile('(XLSX?)')

def collect_sheet_uris(html_file_paths:typing.List[Path]):
    uris = {}
    pattern = re.compile('(XLSX?)')

    for file_path in tqdm.tqdm(html_file_paths, total=len(html_file_paths), desc='HTM(L)'):

        with open(file_path, 'r') as f:
            soup = BeautifulSoup(f, features="html.parser")

        elements = soup.find_all('td', attrs={'data-th':'Dokument'})
        for element in elements:
            title = element.div.p.strong.text.strip()
            href = element.find('a', attrs={'title':RE_SHEET})
            if href is None: continue
            uris[title] = href['href']
    return uris

# Cell
def download_sheet(uri:str, sheet_path:typing.Union[Path,str],
                   verbose:bool=False):
    sheet_path.mkdir(exist_ok=True)
    path = Path(sheet_path) / uri.split('/')[-1]
    with open(path, 'wb') as f:
        r = requests.get(uri)
        if verbose: print(f'Writing to {path}')
        f.write(r.content)

# Cell
def download_multiple_sheets(uris:typing.Dict[str,str], sheet_path:typing.Union[Path,str],
                             t_sleep:float=.01, nmax:int=None):
    file_title_maps = {uri.split('/')[-1]: title for title, uri in uris.items()}
    for i, (title, uri) in tqdm.tqdm(enumerate(uris.items()), desc='File', total=len(uris)):
        if nmax is not None and i > nmax: break
        if (Path(sheet_path) / uri.split('/')[-1]).exists(): continue
        download_sheet(uri, sheet_path=sheet_path)
        time.sleep(t_sleep)
    return file_title_maps

# Cell
RE_FNAME = re.compile('(\.xlsx?)')

# Cell
def is_date(s:str, fun:typing.Callable):
    try:
        _ = fun(s)
        return True
    except:
        return False

VOTE_COLS = ['ja', 'nein', 'Enthaltung', 'ungültig', 'nichtabgegeben']

def get_sheet_df(sheet_file:typing.Union[str,Path], file_title_maps:typing.Dict[str,str]=None):
    'Parsing xlsx and xls files into dataframes'

    if Path(sheet_file).stat().st_size == 0:
        loguru.warning(f'{sheet_file} is of size 0, skipping ...')
        return

    dfs = pd.read_excel(sheet_file, sheet_name=None)

    assert len(dfs) == 1, 'The sheet file has more than one page, that\'s unexpected.'

    for name, df in dfs.items():
        df['sheet_name'] = name

    assert not (df[VOTE_COLS].sum(axis=1) == 0).any()
    assert not (df[VOTE_COLS].sum(axis=1) > 1).any()

    date, title = None, None
    if file_title_maps is not None:
        title, date = handle_title_and_date(file_title_maps[sheet_file.name], sheet_file)

    df['date'] = date
    df['title'] = title

    return df.pipe(disambiguate_party)

def handle_title_and_date(full_title:str, sheet_file):
    'Extracting the title of the roll call vote and the date'
    title = full_title.split(':')
    date = title[0]
    if is_date(date, lambda x: pd.to_datetime(x, dayfirst=True)):
        date = pd.to_datetime(date, dayfirst=True)
        title = ':'.join(title[1:])
    elif is_date(sheet_file.name.split('_')[0], pd.to_datetime):
        date = pd.to_datetime(sheet_file.name.split('_')[0])
        title = full_title
    else:
        date = None
        title = full_title
    return title, date

PARTY_MAP = {'BÜNDNIS`90/DIE GRÜNEN': 'BÜ90/GR', 'DIE LINKE': 'DIE LINKE.', 'fraktionslos': 'Fraktionslos', 'fraktionslose': 'Fraktionslos'}

def disambiguate_party(df:pd.DataFrame, col:str='Fraktion/Gruppe', party_map:dict=None):
    if party_map is None: party_map = PARTY_MAP
    df[col] = df[col].apply(lambda x: x if x not in party_map else party_map[x])
    return df

# Cell
def get_squished_dataframe(df:pd.DataFrame, id_col:str='Bezeichnung',
                           feature_cols:typing.List[str]=VOTE_COLS,
                           other_cols:typing.List=None):

    other_cols = ['date', 'title'] if other_cols is None else other_cols
    tmp = df.loc[:, [id_col] + feature_cols + other_cols]
    tmp['issue'] = df['date'].dt.date.apply(str) + ' ' + df['title']

    tmp = tmp.set_index([id_col, 'issue'] + other_cols)
    tmp = (tmp[tmp == 1].stack()
            .reset_index()
            .drop(0,1)
            .rename(columns={f'level_{2+len(other_cols)}':'vote'}))
    return df.join(tmp.set_index(['Bezeichnung','date','title']), on=['Bezeichnung','date','title']).drop(columns=VOTE_COLS)

# Cell
DTYPES = {'Wahlperiode':int, 'Sitzungnr': int, 'Abstimmnr': int, 'Fraktion/Gruppe':str,
          'Name': str, 'Vorname': str, 'Titel': str, 'vote': str, 'issue': str,
#           'ja': bool, 'nein': bool, 'Enthaltung': bool, 'ungültig': bool, 'nichtabgegeben': bool, 'Bemerkung': str,
          'Bezeichnung': str, 'sheet_name': str,
          'date': 'datetime64[ns]', 'title':str}

def set_sheet_dtypes(df:pd.DataFrame):
    for col, dtype in DTYPES.items():
        df[col] = df[col].astype(dtype)
    return df

# Cell
def get_multiple_sheets_df(sheet_files:typing.List[typing.Union[Path,str]],
                           file_title_maps:typing.Dict[str,str]=None):
    df = []
    for sheet_file in tqdm.tqdm(sheet_files, total=len(sheet_files), desc='Sheets'):
        df.append(
            (get_sheet_df(sheet_file, file_title_maps=file_title_maps)
             .pipe(get_squished_dataframe)
             .pipe(set_sheet_dtypes))
        )
    return pd.concat(df, ignore_index=True)

# Cell
def get_multiple_sheets(html_path, sheet_path, nmax:int=None):
    html_path, sheet_path = Path(html_path), Path(sheet_path)
    html_file_paths = get_file_paths(html_path, pattern=RE_HTM)
    sheet_uris = collect_sheet_uris(html_file_paths)
    file_title_maps = download_multiple_sheets(sheet_uris, sheet_path=sheet_path, nmax=nmax)
    sheet_files = get_file_paths(sheet_path, pattern=RE_FNAME)
    df = get_multiple_sheets_df(sheet_files, file_title_maps=file_title_maps)
    return df


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_similarities.ipynb (unless otherwise specified).

__all__ = ['get_votes_by_party', 'prepare_votes_of_mdb', 'align_mdb_with_parties', 'compute_similarity',
           'get_mdb_party_similarity', 'plot_similarity_over_time', 'align_party_with_party',
           'align_party_with_all_parties', 'get_party_party_similarity']

# Cell
import pandas as pd
import plotly.graph_objects as go
import plotly.express as px
import numpy as np
from scipy import spatial
from bundestag import parsing

pd.options.plotting.backend = "plotly"

# Cell
def get_votes_by_party(df:pd.DataFrame):
    party_votes = (df.groupby(['Fraktion/Gruppe', 'date', 'title'])['vote']
                   .value_counts(normalize=True)
                   .to_frame('fraction')
                   .join(df.groupby(['Fraktion/Gruppe', 'date', 'title'])['vote']
                         .value_counts(normalize=False)
                         .to_frame('# votes')))
    df_plot = party_votes.reset_index()

    index = []
    for party, date, title in df_plot[['Fraktion/Gruppe', 'date', 'title']].drop_duplicates().values:
        index.extend([[party, date, title, vote] for vote in parsing.VOTE_COLS])

    index = np.array(index)
    index = pd.MultiIndex.from_arrays(index.T, names=['Fraktion/Gruppe', 'date', 'title', 'vote'])
    df_plot = (df_plot
               .set_index(['Fraktion/Gruppe', 'date', 'title', 'vote'])
               .reindex(index)
               .fillna(0))

    party_votes = df_plot.pivot_table(index=['Fraktion/Gruppe', 'date', 'title'],
                                  columns='vote', values='fraction')
    return party_votes, df_plot.reset_index()

# Cell
def prepare_votes_of_mdb(df:pd.DataFrame, mdb:str):
    assert mdb in df['Bezeichnung'].unique()
    mask = df['Bezeichnung'] == mdb

    mdb_votes = df.loc[mask, ['date', 'title', 'vote']]
    mdb_votes['vote'] = mdb_votes['vote'].astype('category')
    mdb_votes['vote'].cat.set_categories(parsing.VOTE_COLS, inplace=True)
    # ['ja', 'nein', 'nichtabgegeben', 'Enthaltung']

    mdb_votes = pd.get_dummies(mdb_votes, columns=['vote'], prefix='', prefix_sep='')
    return mdb_votes

# Cell
def align_mdb_with_parties(mdb_votes:pd.DataFrame, party_votes:pd.DataFrame):
    return mdb_votes.join(party_votes.reset_index('Fraktion/Gruppe'), on=['date', 'title'], lsuffix='_mdb', rsuffix='_party')

# Cell
def compute_similarity(df:pd.DataFrame, lsuffix:str, rsuffix:str,
                       similarity_metric=lambda u,v: 1 - spatial.distance.cosine(u,v)):
    lcols = [f'{v}_{lsuffix}' for v in parsing.VOTE_COLS]
    rcols = [f'{v}_{rsuffix}' for v in parsing.VOTE_COLS]
    U = df[lcols].values
    V = df[rcols].values
    df['similarity'] = [similarity_metric(u, v) for u,v in zip(U,V)]
    return df

# Cell
def get_mdb_party_similarity(similarity_mdb_party:pd.DataFrame):
    return similarity_mdb_party.groupby('Fraktion/Gruppe')['similarity'].describe().sort_values('mean')

# Cell
def plot_similarity_over_time(df:pd.DataFrame, grp_col:str, time_bin:str='y', title:str=None):
    y = 'avg. similarity'
    fig = df.groupby([pd.Grouper(key='date', freq=time_bin),grp_col])['similarity'].mean().to_frame(y).reset_index().plot(kind='line', x='date', y=y, color=grp_col, title=title)
    fig.update_layout(xaxis_title=f'Time [{time_bin}]',
                      yaxis_title=f'{y} (0 = dissimilar, 1 = identical)')
    return fig

# Cell
def align_party_with_party(party_votes:pd.DataFrame,
                           party_a:str, party_b:str):
    tmp = party_votes.reset_index(level=0)
    mask_a = tmp['Fraktion/Gruppe'] == party_a
    mask_b = tmp['Fraktion/Gruppe'] == party_b
    return tmp.loc[mask_a].join(tmp.loc[mask_b], lsuffix='_a', rsuffix='_b').reset_index()

    lcols = ['ja_a', 'nein_a', 'nichtabgegeben_a', 'Enthaltung_a']
    rcols = ['ja_b', 'nein_b', 'nichtabgegeben_b', 'Enthaltung_b']
    return tmp.pipe(compute_similarity, lcols=lcols, rcols=rcols)

# Cell
def align_party_with_all_parties(party_votes:pd.DataFrame, party:str):
    similarity_party_party = []
    for party_b in party_votes.index.get_level_values(level='Fraktion/Gruppe').unique():
        if party_b == party: continue
        tmp = align_party_with_party(party_votes, party_a=party, party_b=party_b)
        similarity_party_party.append(tmp)
    similarity_party_party = pd.concat(similarity_party_party, ignore_index=True)
    notna = similarity_party_party['Fraktion/Gruppe_b'].notna()
    return similarity_party_party.loc[notna]

# Cell
def get_party_party_similarity(similarity_party_party:pd.DataFrame):
    return similarity_party_party.groupby('Fraktion/Gruppe_b')['similarity'].describe().sort_values('mean')


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"get_file_paths": "00_html_parsing.ipynb",
         "RE_HTM": "00_html_parsing.ipynb",
         "collect_sheet_uris": "00_html_parsing.ipynb",
         "RE_SHEET": "00_html_parsing.ipynb",
         "download_sheet": "00_html_parsing.ipynb",
         "download_multiple_sheets": "00_html_parsing.ipynb",
         "RE_FNAME": "00_html_parsing.ipynb",
         "is_date": "00_html_parsing.ipynb",
         "get_sheet_df": "00_html_parsing.ipynb",
         "handle_title_and_date": "00_html_parsing.ipynb",
         "disambiguate_party": "00_html_parsing.ipynb",
         "VOTE_COLS": "00_html_parsing.ipynb",
         "PARTY_MAP": "00_html_parsing.ipynb",
         "get_squished_dataframe": "00_html_parsing.ipynb",
         "set_sheet_dtypes": "00_html_parsing.ipynb",
         "DTYPES": "00_html_parsing.ipynb",
         "get_multiple_sheets_df": "00_html_parsing.ipynb",
         "get_multiple_sheets": "00_html_parsing.ipynb",
         "pd.options.plotting.backend": "02_gui.ipynb",
         "get_votes_by_party": "01_similarities.ipynb",
         "prepare_votes_of_mdb": "01_similarities.ipynb",
         "align_mdb_with_parties": "01_similarities.ipynb",
         "compute_similarity": "01_similarities.ipynb",
         "get_mdb_party_similarity": "01_similarities.ipynb",
         "plot_similarity_over_time": "01_similarities.ipynb",
         "align_party_with_party": "01_similarities.ipynb",
         "align_party_with_all_parties": "01_similarities.ipynb",
         "get_party_party_similarity": "01_similarities.ipynb",
         "GUI": "02_gui.ipynb"}

modules = ["parsing.py",
           "similarity.py",
           "gui.py"]

doc_url = "https://eschmidt42.github.io/bundestag/"

git_url = "https://github.com/eschmidt42/bundestag/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_gui.ipynb (unless otherwise specified).

__all__ = ['GUI']

# Cell
try:
    from bundestag import similarity
except ImportError:
    import sys
    sys.path.append('..')
    from bundestag import similarity

import ipywidgets as widgets
import pandas as pd

pd.options.plotting.backend = "plotly"

# Cell
class GUI:
    def __init__(self, df:pd.DataFrame):
        self.df = df
        self.party_votes, self.df_plot = similarity.get_votes_by_party(df)
        self.mdbs = df['Bezeichnung'].unique()
        self.parties = df['Fraktion/Gruppe'].unique()
        self.init_widgets()
    def render(self):
        return widgets.VBox([
            self.mdb_widget,
            self.party_widget,
            self.start_widget,
            self.end_widget,
            self.submit_widget,
            self.selection_widget,
            self.display_widget,
        ])
    def on_click(self, change):
        mdb = self.mdb_widget.value
        party = self.party_widget.value
        assert mdb is not None, 'Please choose an MdB'
        assert party is not None, 'Please choose a Party'
        start_date = self.start_widget.value
        end_date = self.end_widget.value
        self.selection_widget.value = f'Selected: MdB = {mdb}, Party = {party}, date range = {start_date} - {end_date}'

        df, party_votes = self.filter_dfs(start_date, end_date)

        mdb_votes = similarity.prepare_votes_of_mdb(df, mdb)
        similarity_mdb_party = (similarity.align_mdb_with_parties(mdb_votes, party_votes)
                                .pipe(similarity.compute_similarity, lsuffix='mdb', rsuffix='party'))

        similarity_party_party = (similarity.align_party_with_all_parties(party_votes, party)
                                  .pipe(similarity.compute_similarity, lsuffix='a', rsuffix='b'))

        self.display_widget.clear_output()
        with self.display_widget:
            similarity.plot_similarity_over_time(similarity_mdb_party,
                                                 'Fraktion/Gruppe',
                                                 title=f'{mdb} vs time').show()
            similarity.plot_similarity_over_time(similarity_party_party,
                                                 'Fraktion/Gruppe_b',
                                                 title=f'{party} vs time').show()

    def init_widgets(self):
        # prints out the current selection of MdB, party and timespan
        self.selection_widget = widgets.Label()

        # widget to render figures in
        self.display_widget = widgets.Output()

        # submit button
        self.submit_widget = widgets.Button(description='Submit')
        self.submit_widget.on_click(self.on_click)

        # widgets to set MdB, party, timespan
        self.mdb_widget = widgets.Combobox(
            placeholder='Choose an MdB',
            options=tuple(self.mdbs),
            description='MdB:',
            ensure_option=True,
        )
        self.party_widget = widgets.Combobox(
            placeholder='Choose a Party',
            options=tuple(self.parties),
            description='Party:',
            ensure_option=True,
        )
        self.start_widget = widgets.DatePicker(description='Start date')
        self.end_widget = widgets.DatePicker(description='End date')

    def filter_dfs(self, start_date, end_date):
        if (start_date is None or end_date is None):
            return self.df, self.party_votes

        mask = (self.df['date']>=str(start_date)) & (self.df['date']<=str(end_date))
        df = self.df.loc[mask]

        date_range = pd.date_range(start_date, end_date)
        mask = self.party_votes.index.get_level_values('date').isin(date_range)
        party_votes = self.party_votes.loc[mask]
        return df, party_votes